// A client that uses the three utility services to
// track discoveries.

package corejini.chapter19;

import net.jini.lookup.ServiceDiscoveryManager;
import net.jini.event.EventMailbox;
import net.jini.event.MailboxRegistration;
import net.jini.lease.LeaseRenewalService;
import net.jini.lease.LeaseRenewalSet;
import net.jini.discovery.LookupDiscoveryService;
import net.jini.discovery.LookupDiscoveryRegistration;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.lease.LeaseDeniedException;
import java.rmi.MarshalledObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

public class Collector {
    private ServiceDiscoveryManager sdm;
    private LookupDiscoveryService lds;
    private EventMailbox ems;
    private LeaseRenewalService lrs;
    private LeaseRenewalSet leaseSet;
    private RemoteEventListener mboxListener;
    private MailboxRegistration mboxReg;
    // 2 hour run
    private static final long DURATION = 2 * 60 * 60 * 1000;
    
    public Collector() throws IOException {
        sdm = new ServiceDiscoveryManager(null, null);
    }
    
    // Find all of the three utility services
    public void findUtilityServices() 
        throws Exception {
        System.out.println("Finding utility services");
        
        // service items for the three services
        ServiceItem ldsItem, emsItem, lrsItem;
        ServiceTemplate tmpl;
        long waitDur = 1 * 60 * 1000;        // 1 minute

        try {
            Class[] types = new Class[]{ LookupDiscoveryService.class };
            tmpl = new ServiceTemplate(null, types, null);
            System.out.println("Searching for LookupDiscoveryService...");
            if ((ldsItem = sdm.lookup(tmpl, null, waitDur)) == null) {
                throw new Exception("No LookupDiscovery found before timeout");
            } else {
                System.out.println("...found!");
            }
        } catch (InterruptedException ex) {
            throw new Exception("No LookupDiscovery found before timeout");
        }
        
        try {
            Class[] types = new Class[]{ EventMailbox.class };
            tmpl = new ServiceTemplate(null, types, null);
            System.out.println("Searching for EventMailboxService...");
            if ((emsItem = sdm.lookup(tmpl, null, waitDur)) == null) {
                throw new Exception("No EventMailbox found before timeout");
            } else {
                System.out.println("...found!");
            }
        } catch (InterruptedException ex) {
            throw new Exception("No EventMailbox found before timeout");
        }
        
        try {
            Class[] types = new Class[]{ LeaseRenewalService.class };
            tmpl = new ServiceTemplate(null, types, null);
            System.out.println("Searching for LeaseRenewalService...");
            if ((lrsItem = sdm.lookup(tmpl, null, waitDur)) == null) {
                throw new Exception("No LeaseRenewal found before timeout");
            } else {
                System.out.println("...found!");
            }
        } catch (InterruptedException ex) {
            throw new Exception("No LeaseRenewal found before timeout");
        }
        
        // At this point, we've got all the service items.
        // Do some checking to make sure there aren't
        // codebase problems.
        
        if (ldsItem.service == null) {
            throw new Exception("No proxy for LookupDiscovery service. " +
                                "Check for codebase problems?");
        } else {
            lds = (LookupDiscoveryService) ldsItem.service;
        }
        
        if (emsItem.service == null) {
            throw new Exception("No proxy for EventMailbox service. " +
                                "Check for codebase problems?");
        } else {
            ems = (EventMailbox) emsItem.service;
        }
        
        if (lrsItem.service == null) {
            throw new Exception("No proxy for LeaseRenewal service. " +
                                "Check for codebase problems?");
        } else {
            lrs = (LeaseRenewalService) lrsItem.service;
        }
    }
    
    // Create a mailbox registration and record its
    // listener.
    public void setupEventMailbox() 
        throws LeaseDeniedException, RemoteException {
        mboxReg = ems.register(DURATION);
        
        // Store the listener
        mboxListener = mboxReg.getListener();
        
        System.out.println("Event mailbox established");
    }
    
    // create a lease renewal set that we can use to
    // stuff leases into.  also sign up to receive
    // informational events.
    public void setupLeaseRenewals() 
        throws RemoteException {
        leaseSet = lrs.createLeaseRenewalSet(DURATION);
        
        // Have expiration warnings sent to the mailbox
        // 1 minute before the lease set expires
        leaseSet.setExpirationWarningListener(mboxListener, 
                                              1 * 60 * 1000,
                                              null);
        
        // Have any renewal failures also sent to
        // the mailbox.
        leaseSet.setRenewalFailureListener(mboxListener,
                                           null);
        
        // Now, use the lease renewal service to
        // renew the lease on the mailbox for
        // double the duration.  This lets the mailbox
        // stick around a little longer than the
        // rest of the system.
        leaseSet.renewFor(mboxReg.getLease(), 
                          2 * DURATION);
        
        System.out.println("Lease set established");
    }
    
    public void setupDiscovery() 
        throws RemoteException {
        LookupDiscoveryRegistration luReg =
            lds.register(DiscoveryGroupManagement.ALL_GROUPS, 
                         null,         // no locators
                         mboxListener, // Events forwarded to mbox
                         null,         // handback object
                         DURATION);
        
        // Add the lease for the registration to the
        // LeaseSet from the LeaseRenewalService so
        // that it will be renewed.
        leaseSet.renewFor(luReg.getLease(), DURATION);
        
        System.out.println("Discovery registration established");
    }
    
    private MailboxRegistration getMboxRegistration() {
        return mboxReg;
    }
    
    public static void main(String[] args) {
        if (args.length != 1) {
            System.out.println("Usage: <file>");
            System.exit(1);
        }
        
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(
                    new RMISecurityManager());
        }
        
        try {
            File file = new File(args[0]);
            ObjectOutputStream out =
                new ObjectOutputStream(
                    new FileOutputStream(file));

            Collector collector = new Collector();
            
            collector.findUtilityServices();
            
            collector.setupEventMailbox();
            collector.setupLeaseRenewals();
            collector.setupDiscovery();
            
            System.out.println("Services connected.");
            System.out.println("Saving mailbox...");
            
            Object mboxReg = collector.getMboxRegistration();
	    MarshalledObject mobj = new MarshalledObject(mboxReg);
            out.writeObject(mobj);
            out.flush();
            out.close();
            System.out.println("...mailbox successfully saved to " +
                               args[0]);
            
            System.exit(1);
        } catch (Exception ex) {
            System.err.println("Error: " + ex.getMessage());
            ex.printStackTrace();
            System.exit(1);
        }
    }
}

